#include <cstdlib>
#include <stdio.h>
#include <set>
#include <map>
#include <boost/config.hpp>
#include <boost/program_options/detail/config_file.hpp>
#include <boost/program_options/parsers.hpp>

#include <boost/filesystem/operations.hpp>

#include "logger.h"
#include "settings.h"
#include "tools.h"

#define MAX_BUF_SZ 256+128
#define SKIP_WHITE(v) {while (isspace(*(v))) (v)++;}

/* List of accepted options, their values will be put in 
 * this->opt_values */
static const char *opts[] = {
    "download_dir",
    "port_start",
    "port_end",
    "upload_rate_limit",
    "download_rate_limit",
    "upload_slot_limit",
    "refresh_rate",
    "ip_filter",
    "storage_mode",
    "notifications",
    "enable_upnp",
    "enable_natpmp",
    "window_width",
    "window_height",
    "window_x",
    "window_y",
    "action_on_completion",
};

namespace bf = boost::filesystem;

Settings *g_settings;

/** 
 * Create a settings object with initial configuration 
 * loaded from ~/.torium/settings.conf
 **/
Settings::Settings()
{
#ifdef WIN32
    char c[512];

    snprintf(c, 512, "%s%s", getenv("HOMEDRIVE"), getenv("HOMEPATH"));

    this->home_path  = c;
    this->share_path = strdup(c);
#else
    this->home_path = getenv("HOME");
    this->share_path = strdup(TORIUM_SHARE_PATH);

    if (!strlen(this->share_path)) {
        this->share_path = getenv("HOME");
    }
#endif

    this->conf_dir       = this->home_path + DIR_SEP + ".torium";
    this->torrent_dir    = this->conf_dir  + DIR_SEP + "torrents";
    this->fastresume_dir = this->conf_dir  + DIR_SEP + "fastresume";
    this->monitor_dir    = this->conf_dir  + DIR_SEP + "monitor";
    this->states_dir     = this->conf_dir  + DIR_SEP + "states";
    this->pipe_file      = this->conf_dir  + DIR_SEP + ".torium.pipe";
    this->pid_file       = this->conf_dir  + DIR_SEP + ".torium.pid";
    this->conf_file      = this->conf_dir  + DIR_SEP + "settings.conf";

    if (!init()) {
        fprintf(stderr, "Unable to create the necessary directories\n"
                        "Please check the permissions of %s\n",
                        this->home_path.c_str());
        exit(1);
    }
    reset_to_defaults();
    load_config(this->conf_file.c_str());
}

bool
Settings::load_config(const char *path)
{
    char in[MAX_BUF_SZ];
    char *k, *v;
    int  x;
    bf::path p(path);

    if (!bf::exists(p)) {
        g_logger->message("Creating ~/.torium/settings.conf");
        cfg_h.open((this->conf_dir+DIR_SEP+"settings.conf").c_str(),
                std::ios::in | std::ios::out | std::ios::trunc);
        if (cfg_h.fail()) {
            g_logger->error("Could not create ~/.torium/settings.conf", true);
            return false;
        }
        cfg_h.close();

        this->flush_config();
    } else {
        cfg_h.open((this->conf_dir+DIR_SEP+"settings.conf").c_str(),
                   std::ios::in | std::ios::out);
        if (cfg_h.fail()) {
            g_logger->error("Could not open ~/.torium/settings.conf", true);
            return false;
        }

        int line=1;
        char *ke;
        int sz = sizeof(opts)/sizeof(char *);
        while (!cfg_h.eof()) {
            cfg_h.getline(in, MAX_BUF_SZ);
            if (*in == '\0')
                break;
            k = in;
            SKIP_WHITE(k);
            v = k;
            do v++; while (isalnum(*v) || *v == '_' || *v == '-');
            ke = v;
            SKIP_WHITE(v);
            if (*v != '=') {
                g_logger->errorf("syntax error on line %d, missing '='", line);
                line++;
                continue;
            }
            *ke = '\0';
            v++;
            SKIP_WHITE(v);

            unsigned int found = 0;
            for (x=0; x<sz; x++) {
                if (strcmp(opts[x], k) == 0) {
                    this->opt_values[k] = v;
                    found = 1;
                    break;
                }
            }
            if (!found)
                g_logger->warningf("unknown option '%s'", k);

            line++;
        }
        cfg_h.close();
    }

    return true;
}

/** 
 * Attempt to create all necessary directories
 * return false on failure
 **/
bool
Settings::init()
{
    bf::path p(this->conf_dir);

    /* Check if the ~/.torium/ exists, and if not then 
     * attempt to create it */
    if (!bf::exists(p)) {
        g_logger->errorf("Creating %s", p.string().c_str());
        if (!bf::create_directory(p)) {
            g_logger->errorf("Could not create %s", p.string().c_str());
            return false;
        }
    }
    /* Make sure it is a directory and not a file */
    if (!bf::is_directory(p)) {
        g_logger->errorf("%s is not a directory", p.string().c_str());
        return false;
    }

    p = bf::path(this->torrent_dir);

    if (!bf::exists(p)) {
        g_logger->errorf("Creating %s", p.string().c_str());
        if (!bf::create_directory(p)) {
            g_logger->errorf("Could not create %s", p.string().c_str());
            return false;
        }
    }
    if (!bf::is_directory(p)) {
        g_logger->errorf("%s is not a directory", p.string().c_str());
        return false;
    }

    p = bf::path(this->fastresume_dir);

    if (!bf::exists(p)) {
        g_logger->errorf("Creating %s", p.string().c_str());
        if (!bf::create_directory(p)) {
            g_logger->errorf("Could not create %s", p.string().c_str());
            return false;
        }
    }
    if (!bf::is_directory(p)) {
        g_logger->errorf("%s is not a directory", p.string().c_str());
        return false;
    }

    p = bf::path(this->monitor_dir);

    if (!bf::exists(p)) {
        g_logger->errorf("Creating %s", p.string().c_str());
        if (!bf::create_directory(p)) {
            g_logger->errorf("Could not create %s", p.string().c_str());
            return false;
        }
    }
    if (!bf::is_directory(p)) {
        g_logger->errorf("%s is not a directory", p.string().c_str());
        return false;
    }

    p = bf::path(this->states_dir);

    if (!bf::exists(p)) {
        g_logger->errorf("Creating %s", p.string().c_str());
        if (!bf::create_directory(p)) {
            g_logger->errorf("Could not create %s", p.string().c_str());
            return false;
        }
    }
    if (!bf::is_directory(p)) {
        g_logger->errorf("%s is not a directory", p.string().c_str());
        return false;
    }

    return true;
}

/** 
 * Fill in opt_values with default values
 **/
void
Settings::reset_to_defaults()
{
    this->opt_values["download_dir"]        = this->home_path+DIR_SEP+"Downloads";
    this->opt_values["port_start"]          = "6000";
    this->opt_values["port_end"]            = "6020";
    this->opt_values["upload_rate_limit"]   = "-1";
    this->opt_values["download_rate_limit"] = "-1";
    this->opt_values["upload_slot_limit"]   = "15";
    this->opt_values["refresh_rate"]        = "1000";
    this->opt_values["ip_filter"]           = "yes";
    this->opt_values["storage_mode"]        = "sparse";
    this->opt_values["notifications"]       = "yes";
    this->opt_values["enable_upnp"]         = "yes";
    this->opt_values["enable_natpmp"]       = "yes";
    this->opt_values["window_width"]        = "730";
    this->opt_values["window_height"]       = "180";
    this->opt_values["window_x"]            = "50";
    this->opt_values["window_y"]            = "50";
    this->opt_values["action_on_completion"] = "nothing";
}

/** 
 * Save the current settings to the users settings.conf
 **/
bool
Settings::flush_config()
{
    int x=0;
    std::fstream cfg_h((this->conf_dir+DIR_SEP+"settings.conf").c_str(),
                       std::ios::in | std::ios::out | std::ios::trunc);
    /* Loop through all options in opt_values and 
     * simply print them each as "var = value" to the file. */
    for (std::map<std::string, std::string>::const_iterator i = this->opt_values.begin();
            i != this->opt_values.end(); i++) {
        x++;
        std::string out = (i->first + " = " + i->second + "\n");
        cfg_h.write(out.c_str(), out.length());
        if (cfg_h.fail()) {
            g_logger->error("could not write to settings.conf", true);
            return false;
        }
    }

    cfg_h.close();

    g_logger->messagef("Saved %d option(s) to ~/.torium/settings.conf", x);
    return true;
}

Settings::~Settings()
{
    this->flush_config();
    free(this->share_path);
}

/**
 * Set the specified option to the given value
 **/
bool
Settings::set_option(const char *option, const char *value)
{
    this->opt_values[option] = value;

    return true;
}

bool
Settings::set_option(const char *option, const int value)
{
    char c[128];
    sprintf(c, "%d", value);

    return set_option(option, c);
}

bool
Settings::set_option(const char *option, const std::string value)
{
    return set_option(option, value.c_str());
}

/**
 * Return the option's value or NULL if it is not set
 **/
const char *
Settings::get_option(const char *option)
{
    std::map<std::string, std::string>::iterator x = this->opt_values.find(option);
    if (x == this->opt_values.end()) {
        g_logger->warningf("no value set for option `%s'", option);
        return "";
    } else
        return x->second.c_str();
}
