#include "tools.h"
#include "logger.h"
#include "settings.h"

const char *Priorities_str[] = {
    "skip",
    "normal",
    "fairly high",
    "pretty high",
    "high",
    "high",
    "max",
    "supermax"
};

const char *States_str[] = {
    "Queued",
    "Checking",
    "Downloading",
    "Metadata",
    "Downloading",
    "Finished",
    "Seeding",
    "Allocating",
    "Paused",
};

/** 
 * URLEncodes the given string
 * example:
 * urlencode("value=a string")
 * gives
 * value=a%20string
 **/
std::string
urldecode(std::string url)
{
    std::string buffer = "";
    int len = url.length();

    for(int i = 0; i < len; i++) {
        int j = i;
        char ch = url.at(j);
        if(ch == '%') {
            char tmpstr[] = "0x0__";
            int chnum;
            tmpstr[3] = url.at(j+1);
            tmpstr[4] = url.at(j+2);
            chnum = strtol(tmpstr, NULL, 16);
            buffer += chnum;
            i += 2;
        } else {
            buffer += ch;
        }
    }
    return buffer;
}

/** 
 * Converts a float to the lowest possible "abbreviation"
 **/
const char*
fix_rate(float val)
{
    char *ret = (char*)malloc(32);
    const char* prefix[] = {"kB/s", "MB/s", "GB/s", "TB/s", "PB/s"};
    const int num = sizeof(prefix) / sizeof(const char*);

    if(val < 1000.f) {
        val /= 1000.f;
        sprintf(ret, "%.1f Kb/s", val);

        return ret;
    }

    for(int i = 0; i < num; ++i) {
        val /= 1000.f;
        if(fabs(val) < 1000.f) {
            sprintf(ret, "%.1f %s", val, prefix[i]);

            return ret;
        }
    }

    sprintf(ret, "%.1f ??", val);

    return ret;
}

/** 
 * Converts downspeed, total file size and the amount downloaded into
 * an estimate of how much time reimains in seconds
 **/
unsigned
get_eta(float dspeed, long total_size, long downloaded)
{
    return (unsigned)((total_size-downloaded)/dspeed);
}

/** 
 * Converts seconds to the appropriate timeformat with a maximum
 * of two timetypes
 * example:
 * fix_time(3660)
 * gives
 * 1h1m
 **/
std::string
fix_time(unsigned s, int args)
{
    unsigned days, hours, minutes, seconds;
    char c[64];

    if(s == 0 || s > 86400*92)
        return "N/A";

    days = hours = minutes = seconds = 0;

    while(s >= 86400) {
        s -= 86400;
        days++;
    }
    while(s >= 3600) {
        s -= 3600;
        hours++;
    }
    while(s >= 60) {
        s -= 60;
        minutes++;
    }
    seconds = s;

    if(days > 0) {
        if(hours > 0)
            sprintf(c, "%dd%dh", days, hours);
        else
            sprintf(c, "%dd", days);
    } else if(hours > 0) {
        if(minutes > 0)
            sprintf(c, "%dh%dm", hours, minutes);
        else
            sprintf(c, "%dh", hours);
    } else if(minutes > 0) {
        if(seconds > 0)
            sprintf(c, "%dm%ds", minutes, seconds);
        else
            sprintf(c, "%dm", minutes);
    } else if (seconds > 0)
        sprintf(c, "%ds", seconds);
    
    return c;
}

const char *
prio_name(const unsigned int prio)
{
    if (prio < PRIO_END)
        return Priorities_str[prio];
    
    return "Unknown";
}

const char *
state_name(const unsigned int state)
{
    if (state < STATE_END)
        return States_str[state];

    return "Unknown";
}

/** 
 * Truncates the given string to the given length
 **/
std::string
truncate(std::string str, unsigned int len)
{
    if(str.length() < len)
        return str;
    else
        return str.substr(0, len)+" ...";
}

/** 
 * Opens a file and determines its filesize
 **/
long
filesize(const std::string path)
{
    long filesize;
    FILE *fh;
    fh = fopen(path.c_str(), "r");

    if (fh == NULL) {
        g_logger->errorf("error opening file %s", path.c_str());
        return 0;
    }

    fseek(fh, 0, SEEK_END);
    filesize = ftell(fh);
    fclose(fh);

    return filesize;
}

/** 
 * Opens Torium's pipe and append a string to it
 **/
bool
pipe_append(const char *str)
{
    FILE *stream;
#ifdef DEBUG
    g_logger->messagef("Appending %s to pipe %s",
                       str, g_settings->pipe_file.c_str());
#endif
    stream = fopen(g_settings->pipe_file.c_str(), "a");
    fprintf(stream, "%s ", str);
    fclose(stream);

    return true;
}
