#include "torrent.h"
#include "libtorrent/session.hpp"
#include <iterator>
#include <fstream>
#include <libtorrent/bencode.hpp>
#include <libtorrent/file_pool.hpp>
#include <boost/filesystem/fstream.hpp>

Torrent *g_torrent;

Torrent::Torrent()
{
    handles_t handles;
    ses = new session();

    bool enable_upnp, enable_natpmp, enable_lsd;
    session_settings ses_settings;
    ses_settings.user_agent = PACKAGE_NAME "/" PACKAGE_VERSION;

    ses_settings.urlseed_wait_retry = 0;

    enable_upnp = (bool)(strcmp(g_settings->get_option("enable_upnp"), "yes") == 0);
    enable_natpmp = (bool)(strcmp(g_settings->get_option("enable_natpmp"), "yes") == 0);
    enable_lsd = true;

    bf::ifstream dht_state_file(".dht_state", std::ios_base::binary);
    dht_state_file.unsetf(std::ios_base::skipws);
    entry dht_state;
    try {
        dht_state = bdecode(
                std::istream_iterator<char>(dht_state_file),
                std::istream_iterator<char>());
    } catch(std::exception&) {}
    ses->start_dht(dht_state);
    ses->add_dht_router(std::make_pair(std::string("router.bittorrent.com"), 6881));
    ses->add_dht_router(std::make_pair(std::string("router.utorrent.com"), 6881));
    ses->add_dht_router(std::make_pair(std::string("router.bitcomet.com"), 6881));
    ses->set_settings(ses_settings);
    dht_state_file.close();

    ses->listen_on(std::make_pair(atoi(g_settings->get_option("port_start")),
                                  atoi(g_settings->get_option("port_end"))));
    ses->set_max_uploads(atoi(g_settings->get_option("upload_slot_limit")));
    ses->set_upload_rate_limit(atoi(g_settings->get_option("upload_rate_limit")));
    ses->set_download_rate_limit(atoi(g_settings->get_option("download_rate_limit")));

    ses->set_severity_level(alert::info);
    if(enable_upnp)     ses->start_upnp();
    if(enable_natpmp)   ses->start_natpmp();
    if(enable_lsd)      ses->start_lsd();

    fastresume_dir = g_settings->fastresume_dir;
    global_uprate = -1;
}

Torrent::~Torrent()
{
    save_fastresume();
}

/** 
 * Adds a torrent
 * Returns the torrent name or "" in case of error
 **/
const char*
Torrent::add_torrent(const std::string path, const std::string download_path)
{
    bf::path save_path;
    bf::path tor_path(path);
    bf::path fastresume_dir(g_settings->fastresume_dir);

    if (path.length() < 3) {
        g_logger->error("Path way too short", true);
        return "";
    }

    if (!bf::exists(tor_path)) {
        g_logger->error("Torrent path does not exist.");
        return "";
    }
    const char *basepath = strrchr(path.c_str(), DIR_SEP);
    if (!*basepath) {
        g_logger->error("something happened");
        return "";
    }

    std::string tmp(basepath);
    std::string tmp2(g_settings->torrent_dir + tmp);
    bf::path new_tor_path(tmp2);

    if (!bf::exists(new_tor_path)) {
        bf::copy_file(tor_path, new_tor_path);
        if (!bf::exists(new_tor_path)) {
            g_logger->error("Could not copy torrent");
            return "";
        }
    }

    std::ifstream in(new_tor_path.string().c_str(), std::ios_base::binary);
    in.unsetf(std::ios_base::skipws);
    entry e;
    try {
        e = bdecode(std::istream_iterator<char>(in),
                          std::istream_iterator<char>());
    } catch (std::exception &ex) {
        g_logger->errorf("torrent::add_torrent(%s): %s",
                         new_tor_path.string().c_str(), ex.what());

        /* There's no reason to save a corrupt torrent file */
        bf::remove(new_tor_path);
        bf::remove(tor_path);
        return "";
    }

    boost::intrusive_ptr<torrent_info> t;
    try {
        t = new torrent_info(e);
    } catch (std::exception &ex) {
        g_logger->errorf("torrent::add_torrent(%s): %s",
                         new_tor_path.string().c_str(), ex.what());

        /* There's no reason to save a corrupt torrent file */
        bf::remove(new_tor_path);
        bf::remove(tor_path);
        return "";
    }

    if (download_path.empty()) {
        save_path = g_settings->get_option("download_dir");
        int ret;
        char *path = (char*)malloc(2+g_settings->states_dir.length()+t->name().length());
        ret = sprintf(path, "%s%c%s", g_settings->states_dir.c_str(), DIR_SEP,
                t->name().c_str());
#ifndef WIN32
        if (ret > 0 && access(path, F_OK) == 0) {
#else
        if (ret > 0) {
#endif      
            std::fstream fh(path, std::ios::in);

            char in[256];
            while (!fh.eof()) {
                fh.getline(in, 256);
                if (*in == '\0')
                    break;

                g_logger->messagef("in: %s", in);
                bf::path p(in);

                if (bf::exists(p) && bf::is_directory(p)) {
                    save_path = p;
                }
            }
            fh.close();

            free(path);
        }
    } else {
        save_path = download_path;

        char *p = (char*)malloc(2+g_settings->states_dir.length()+t->name().length());
        sprintf(p, "%s%c%s", g_settings->states_dir.c_str(), DIR_SEP,
                t->name().c_str());

        std::ofstream fh(p);
        fh << save_path;
        fh.close();

        free(p);
    }

#ifdef DEBUG
    g_logger->messagef("add_torrent(): Download path: '%s'",
                       save_path.string().c_str());
#endif

    if (!bf::exists(save_path)) {
        g_logger->errorf("%s does not exist, attempt to create it",
                         save_path.string().c_str());
        if (!bf::create_directory(save_path)) {
            g_logger->errorf("unable to create %s",
                             save_path.string().c_str());
            return "";
        }
    }
    if (!bf::is_directory(save_path)) {
        g_logger->errorf("%s is not a directory", save_path.string().c_str());
        return "";
    }

    entry resume_data;
    if (!bf::is_directory(fastresume_dir)) {
        g_logger->error(g_settings->fastresume_dir + " is not a directory!");
    } else {
        char *p = (char*)malloc(13+g_settings->fastresume_dir.length()+t->name().length());
        int ret;
        ret = sprintf(p, "%s%c%s.fastresume", g_settings->fastresume_dir.c_str(),
                DIR_SEP, t->name().c_str());
        if (ret <= 0) {
            fprintf(stderr, "error\n");
            abort();
        }
        bf::path fastresume_path(p);

        if (bf::exists(fastresume_path)) {
            bf::ifstream resume_file(fastresume_path, std::ios_base::binary);
            resume_file.unsetf(std::ios::skipws);
            try {
                resume_data = bdecode(
                        std::istream_iterator<char>(resume_file)
                        ,std::istream_iterator<char>());
            } catch (std::exception &ex) {
                g_logger->errorf("fastresume exception caught: %s", ex.what());
                /* There's something wrong with the fastresume, if we simply
                 * resume adding the torrent it will check the file integrity */
            }
        }
    }

    /* Attempt to get torrent state */
    int  r;
    bool pause_state = false;
    char p[512];

    r = snprintf(p, 512, "%s%cP_%s", g_settings->states_dir.c_str(), DIR_SEP,
                        t->name().c_str());
    if (r <= 0) {
        fprintf(stderr, "error\n");
        abort();
    }

    if (bf::exists(bf::path(p))) {
        pause_state = true;
    }

    torrent_handle h;

    try {
        const char *smode = g_settings->get_option("storage_mode");
        if (strcmp(smode, "sparse") == 0) {
            h = ses->add_torrent(t, save_path, resume_data,
                    storage_mode_sparse, pause_state);
        } else if (strcmp(smode, "allocate") == 0) {
            h = ses->add_torrent(t, save_path, resume_data,
                    storage_mode_allocate, pause_state);
        } else if (strcmp(smode, "compact") == 0) {
            h = ses->add_torrent(t, save_path, resume_data,
                    storage_mode_compact, pause_state);
        } else {
            g_logger->warning("torrent::add_torrent(): no proper storage mode set, using sparse");
            g_settings->set_option("storage_mode", "sparse");
            h = ses->add_torrent(t, save_path, resume_data,
                    storage_mode_sparse, pause_state);
        }
    } catch(std::exception &ex) {
        /* Torrent already exists in session */
        g_logger->errorf("torrent::add_torrent(%s): %s",
                         new_tor_path.string().c_str(), ex.what());

        return "";
    }

    h.set_max_connections(60);
    h.set_max_uploads(-1);
    h.set_ratio(0);
    h.resolve_countries(true);

    handles.insert(std::make_pair(path, h));

    return t->name().c_str();
}

bool
Torrent::remove_torrent(const std::string name)
{
    for (handles_t::iterator i = handles.begin();
            i != handles.end(); ++i) {
        torrent_handle &h = i->second;
        if (!h.is_valid()) {
            continue;
        }
        if (h.name() == name) {
            char c[512];
            snprintf(c, 512, "%s%cP_%s", g_settings->states_dir.c_str(), DIR_SEP,
                     h.name().c_str());
            bf::path pause_path(c);

            bf::remove(pause_path);

            ses->remove_torrent(h);
            return true;
        }
    }

    return false;
}

bool
Torrent::get_torrent(const std::string name, torrent_handle &h)
{
    for (handles_t::iterator i = handles.begin();
            i != handles.end(); ++i) {
        torrent_handle &nh = i->second;
        if (!nh.is_valid() || !nh.has_metadata()) {
            continue;
        }
        if (nh.name() == name) {
            h = nh;
            return true;
        }
    }

    g_logger->error("no torrent_info found, no metadata?");
    return false;
}

bool
Torrent::set_ip_filter(ip_filter const &filter)
{
    ses->set_ip_filter(filter);

    return true;
}

bool
Torrent::resume(torrent_handle *h)
{
    g_logger->messagef("Resumed %s", h->name().c_str());
    char c[512];

    h->resume();
    snprintf(c, 512, "%s%cP_%s", g_settings->states_dir.c_str(), DIR_SEP,
             h->name().c_str());
    bf::path pause_path(c);

    return bf::remove(pause_path);
}

bool
Torrent::pause(torrent_handle *h)
{
    char c[512];
    g_logger->messagef("Paused %s", h->name().c_str());
    h->pause();
    
    snprintf(c, 512, "%s%cP_%s", g_settings->states_dir.c_str(), DIR_SEP,
             h->name().c_str());
    bf::path pause_path(c);

    std::ofstream file(pause_path.string().c_str());
    file.close();

    return bf::exists(pause_path);
}

/**
 * Toggle torrent status between paused and resumed
 **/
int
Torrent::toggle_pause(const std::string name)
{
    torrent_handle h;
    get_torrent(name, h);

    if (h.is_paused()) {
        resume(&h);
        return 1;
    } else {
        pause(&h);
        return 0;
    }
}

/**
 * Force torrent file integrity check
 **/
void
Torrent::force_recheck(const std::string name)
{
#if LIBTORRENT_VERSION_MINOR >= 14
    torrent_handle h;
    get_torrent(name, h);
    h.force_recheck();
#else
    g_logger->warning("you need libtorrent >= 0.14 for force_recheck");
#endif
}

/**
 * Force tracker reannounce
 **/
void
Torrent::force_reannounce(const std::string name)
{
    torrent_handle h;
    get_torrent(name, h);
    h.force_reannounce();
}

/**
 * Return a list of all torrents in the current session
 **/
std::vector<torrent_handle>
Torrent::get_torrents()
{
    return ses->get_torrents();
}

/**
 * Loop through all torrents in the session and save their fastresume data
 **/
void
Torrent::save_fastresume()
{
    if (!bf::is_directory(fastresume_dir)) {
        return;
    }

    for (handles_t::iterator i = handles.begin(); i != handles.end(); ++i) {
        torrent_handle& h = i->second;
        if (!h.is_valid() || !h.has_metadata()) {
            continue;
        }

        /* if the torrent is currently checking the file integrity,
         * there is no reason to save a fastresume */
        int state = h.status().state;
        int percentage = (int)(h.status().progress*100);
        if (state == STATE_CHECKING || percentage < 1) {
            continue;
        }

        /* Pause the torrent before attempting to write any resume data */
        h.pause();
        entry data = h.write_resume_data();

        char c[512];
        snprintf(c, 512, "%s%c%s.fastresume", fastresume_dir.file_string().c_str(),
                 DIR_SEP, h.get_torrent_info().name().c_str());
        bf::path fastresume_path(c);

        bf::ofstream out(fastresume_path, std::ios_base::binary);
        out.unsetf(std::ios_base::skipws);
        bencode(std::ostream_iterator<char>(out), data);

        out.close();
    }
}

/**
 * Set a global upload rate limit for all torrents
 **/
void
Torrent::set_global_uprate(int uprate)
{
    this->global_uprate = uprate;
    for (handles_t::iterator i = handles.begin(); i != handles.end(); ++i) {
        torrent_handle& h = i->second;
        if (!h.is_valid()) {
            continue;
        }

        h.set_upload_limit(uprate);
    }
}

/** 
 * Add files to a torrent
 * TODO: move recurse into gui_create_torrent
 **/
void
Torrent::add_files(torrent_info &t,
                   bf::path const &p,
                   bf::path const &l)
{
#if LIBTORRENT_VERSION_MINOR <= 13
    if (l.leaf()[0] == '.') return;
    bf::path f(p/l);

    /* Recurse through all directories */
    if (bf::is_directory(f)) {
        for (bf::directory_iterator i(f), end; i != end; ++i)
            add_files(t, p, l/i->leaf());
    } else {
        g_logger->messagef("Adding %s", l.string().c_str());
        t.add_file(l, bf::file_size(f));
    }
#else
    g_logger->warning("Why are you attempting to use this function?");
#endif
}

/**
 * Create a torrent file
 * TODO: Add support for libtorrent >= 0.14
 **/
void
Torrent::create_torrent(std::vector<Glib::ustring> files,
                        std::vector<Glib::ustring> trackers,
                        const char *path)
{
    const int files_sz    = files.size();
    const int trackers_sz = trackers.size();

    if (!files_sz)
        throw insufficent_files();
    if (!trackers_sz)
        throw insufficent_trackers();
    if (strlen(path) < 1)
        throw invalid_path();

    std::ofstream out(bf::complete(bf::path(path)).string().c_str(),
                      std::ios_base::binary);
#if LIBTORRENT_VERSION_MINOR >= 14
    /*
    libtorrent::file_storage fs;

    for (i = files.begin(); i != files.end(); ++i) {
        libtorrent::add_files(fs, *i);
    }

    for (i = trackers.begin(); i != trackers.end(); ++i) {
        t.add_tracker(*i);
    }

    libtorrent::set_piece_hashes(t, ".");
    libtorrent::bencode(std::ostream_iterator<char>(out), t.generate());
    */
#else
    boost::intrusive_ptr<torrent_info> t(new torrent_info);
    std::vector<Glib::ustring>::iterator i;
    unsigned int n, x;
    unsigned int piece_sz = 256*1024;
    libtorrent::file_pool fp;


    try {
        bf::path full_path;
        size_type full_path_len;
        for (i = files.begin(); i != files.end(); ++i) {
            bf::path p(bf::complete(bf::path(*i)));
            size_type p_len;

            add_files(*t, p.branch_path(), p.leaf());

            full_path_len = full_path.string().length();
            p_len = p.branch_path().string().length();
            if (!full_path_len || full_path_len < p_len) {
                full_path = bf::complete(p.branch_path());
                g_logger->messagef("full path is now %s", full_path.string().c_str());
            }
        }
        t->set_piece_size(piece_sz);

        boost::scoped_ptr<storage_interface> st(
                default_storage_constructor(t, full_path, fp));

        for (i = trackers.begin(); i != trackers.end(); ++i) {
            t->add_tracker(*i);
        }

        n = t->num_pieces();
        std::vector<char> buf(piece_sz);
        for (x = 0; x < n; ++x) {
            st->read(&buf[0], x, 0, t->piece_size(x));
            hasher h(&buf[0], t->piece_size(x));
            t->set_hash(x, h.final());
#ifdef DEBUG
            g_logger->messagef("hash %d/%d", x+1, n);
#endif
        }

        std::string creator(PACKAGE_NAME "/" PACKAGE_VERSION);
        t->set_creator(creator.c_str());

        entry e = t->create_torrent();
        libtorrent::bencode(std::ostream_iterator<char>(out), e);

        g_logger->messagef("Successfully created %s with %d files and %d trackers",
                           path, files_sz, trackers_sz);
    } catch (std::exception &ex) {
        g_logger->errorf("torrent::create_torrent(): %s", ex.what());
    }
#endif
    out.close();
}

const size_type
Torrent::get_total_downloaded()
{
    return ses->status().total_download/1000/1000;
}

bool
Torrent::has_downloaded()
{
    return ses->status().total_download > 0;
}

const size_type
Torrent::get_total_uploaded()
{
    return ses->status().total_upload/1000/1000;
}

const float
Torrent::download_rate()
{
    return ses->status().download_rate;
}

const float
Torrent::upload_rate()
{
    return ses->status().upload_rate;
}

void
Torrent::update_rates()
{
    ses->set_download_rate_limit(atoi(g_settings->get_option("download_rate_limit")));
    ses->set_upload_rate_limit(atoi(g_settings->get_option("upload_rate_limit")));
}

std::auto_ptr<alert>
Torrent::pop_alert()
{
    return ses->pop_alert();
}

void
Torrent::get_num_torrents(int *torrents, int *completed_torrents)
{
    for (handles_t::iterator i = handles.begin();
            i != handles.end(); ++i) {
        torrent_handle h = i->second;
        if (!h.is_valid()) continue;

        ++(*torrents);

        if (h.is_seed()) ++(*completed_torrents);
    }
}

unsigned
Torrent::get_total_eta()
{
    unsigned ret = 0;
    for (handles_t::iterator i = handles.begin();
            i != handles.end(); ++i) {
        torrent_handle h = i->second;
        if (!h.is_valid()) continue;
        torrent_status s = h.status();
        torrent_info   t = h.get_torrent_info();

        ret += get_eta(s.download_rate, t.total_size(), s.total_done);
    }

    return ret;
}
