#include "window.h"
#include "logger.h"

#include <string>
#include <fstream>
#include <gtkmm/stock.h>

MainWindow::MainWindow(std::vector<const char*> torrents, int tmp_x, int tmp_y,
                       int tmp_width, int tmp_height)
{
    int win_width, win_height,
        win_x, win_y;

    /* Initialize status icon (tray icon) */
    status_icon = Gtk::StatusIcon::create(Gtk::Stock::NETWORK);
    status_icon->signal_activate().connect(sigc::mem_fun(*this, &MainWindow::on_status_icon_click));
    status_icon->signal_popup_menu().connect(sigc::mem_fun(*this, &MainWindow::on_status_icon_popup));

    /* Populate the status icon popup menu */
    Gtk::Menu::MenuList &ml_status_icon_menu = status_icon_menu.items();

    ml_status_icon_menu.push_back(Gtk::Menu_Helpers::MenuElem(
        "_Quit",
        sigc::mem_fun(*this, &MainWindow::on_menu_file_quit)));

    status_icon_menu.accelerate(*this);

    set_title("Torium");

    if (tmp_x == -1 || tmp_x == -1) {
        win_x = atoi(g_settings->get_option("window_x"));
        win_y = atoi(g_settings->get_option("window_y"));
    } else {
        win_x = tmp_x;
        win_y = tmp_y;
    }

    if (tmp_width == -1 || tmp_height == -1) {
        win_width  = atoi(g_settings->get_option("window_width"));
        win_height = atoi(g_settings->get_option("window_height"));
    } else {
        win_width  = tmp_width;
        win_height = tmp_height;
    }

    set_size_request(400, 100);
    resize(win_width, win_height);
    move(win_x, win_y);

    /* Main vertical box, this is where everything is located */
    add(m_Box);

    m_refActionGroup = Gtk::ActionGroup::create();

    m_refActionGroup->add(Gtk::Action::create("FileMenu", "_File"));
    m_refActionGroup->add(
            Gtk::Action::create(
                "FileAddTorrent",
                Gtk::Stock::ADD,
                "_Add torrent..."),
            Gtk::AccelKey("<control>O"),
            sigc::mem_fun(*this, &MainWindow::on_menu_file_add_torrent));

    m_refActionGroup->add(
            Gtk::Action::create(
                "FileCreateTorrent",
                Gtk::Stock::NEW,
                "_Create torrent..."),
            Gtk::AccelKey("<control><alt>C"),
            sigc::mem_fun(*this, &MainWindow::on_menu_file_create_torrent));

    m_refActionGroup->add(
            Gtk::Action::create(
                "FileQuit",
                Gtk::Stock::QUIT,
                "_Quit"),
            Gtk::AccelKey("<control>Q"),
            sigc::mem_fun(*this, &MainWindow::on_menu_file_quit));

    m_refActionGroup->add(Gtk::Action::create("EditMenu", "_Edit"));
    m_refActionGroup->add(
            Gtk::Action::create(
                "EditPreferences",
                Gtk::Stock::PREFERENCES,
                "_Preferences..."),
            Gtk::AccelKey("F3"),
            sigc::mem_fun(*this, &MainWindow::on_menu_edit_preferences));

    m_refActionGroup->add(
            Gtk::Action::create(
                "EditLoadIpFilters",
                Gtk::Stock::REFRESH,
                "_Load ip filters"),
            sigc::mem_fun(*this, &MainWindow::on_menu_edit_load_ip_filters));

    m_refActionGroup->add(Gtk::Action::create("ViewMenu", "_View"));
    m_refActionGroup->add(
            Gtk::Action::create(
                "ExpandAll",
                "_Expand all"),
            Gtk::AccelKey("<control>X"),
            sigc::mem_fun(treeview, &TreeView::expand_all));
    m_refActionGroup->add(
            Gtk::Action::create(
                "CollapseAll",
                "_Collapse all"),
            Gtk::AccelKey("<control>C"),
            sigc::mem_fun(treeview, &TreeView::collapse_all));

    m_refActionGroup->add(Gtk::Action::create("TorrentMenu", "_Torrent"));
    m_refActionGroup->add(
            Gtk::Action::create(
                "RemoveTorrent",
#ifndef WIN32
                Gtk::Stock::DELETE,
#else
                Gtk::Stock::CANCEL,
#endif
                "_Remove"),
            Gtk::AccelKey("Delete"),
            sigc::mem_fun(treeview, &TreeView::on_popup_remove_torrent));
    m_refActionGroup->add(
            Gtk::Action::create(
                "PauseResumeTorrent",
                Gtk::Stock::MEDIA_PAUSE,
                "_Pause/Resume"),
            sigc::mem_fun(treeview, &TreeView::on_popup_pauseresume_torrent));
    m_refActionGroup->add(
            Gtk::Action::create(
                "DownloadDir",
                Gtk::Stock::DIRECTORY,
                "Go to download directory"),
            sigc::mem_fun(treeview, &TreeView::on_popup_goto_download_dir));
#if LIBTORRENT_VERSION_MINOR >= 14
    m_refActionGroup->add(
            Gtk::Action::create(
                "ForceRecheck",
                Gtk::Stock::REFRESH,
                "F_orce recheck"),
            sigc::mem_fun(treeview, &TreeView::on_popup_recheck_torrent));
#endif
    m_refActionGroup->add(
            Gtk::Action::create(
                "ForceReannounce",
                Gtk::Stock::REFRESH,
                "_Force reannounce"),
            sigc::mem_fun(treeview, &TreeView::on_popup_reannounce_torrent));
    m_refActionGroup->add(
            Gtk::Action::create(
                "Properties",
                "_Properties"),
            sigc::mem_fun(treeview, &TreeView::on_popup_properties));

    m_refActionGroup->add(Gtk::Action::create("HelpMenu", "_Help"));
    m_refActionGroup->add(
            Gtk::Action::create(
                "Documentation",
                Gtk::Stock::HELP,
                "_Documentation"),
            Gtk::AccelKey("F1"),
            sigc::mem_fun(*this, &MainWindow::on_menu_help_contents));
    m_refActionGroup->add(
            Gtk::Action::create(
                "About",
                Gtk::Stock::ABOUT,
                "_About"),
            Gtk::AccelKey("F2"),
            sigc::mem_fun(*this, &MainWindow::on_menu_help_about));
    m_refActionGroup->add(
            Gtk::Action::create(
                "Stats",
                Gtk::Stock::INFO,
                "_Stats"),
            Gtk::AccelKey("F4"),
            sigc::mem_fun(*this, &MainWindow::on_menu_help_stats));

    m_refUIManager = Gtk::UIManager::create();
    m_refUIManager->insert_action_group(m_refActionGroup);

    add_accel_group(m_refUIManager->get_accel_group());

    Glib::ustring ui_info =
        "<ui>"
        "  <menubar name='MenuBar'>"
        "    <menu action='FileMenu'>"
        "      <menuitem action='FileAddTorrent'/>"
        "      <menuitem action='FileCreateTorrent'/>"
        "      <menuitem action='FileQuit'/>"
        "    </menu>"
        "    <menu action='EditMenu'>"
        "      <menuitem action='EditPreferences'/>"
        "      <menuitem action='EditLoadIpFilters'/>"
        "    </menu>"
        "    <menu action='ViewMenu'>"
        "      <menuitem action='ExpandAll'/>"
        "      <menuitem action='CollapseAll'/>"
        "    </menu>"
        "    <menu action='TorrentMenu'>"
        "      <menuitem action='Properties'/>"
        "      <menuitem action='PauseResumeTorrent'/>"
#if LIBTORRENT_VERSION_MINOR >= 14
        "      <menuitem action='ForceRecheck'/>"
#endif
        "      <menuitem action='ForceReannounce'/>"
        "      <menuitem action='DownloadDir'/>"
        "      <menuitem action='RemoveTorrent'/>"
        "    </menu>"
        "    <menu action='HelpMenu'>"
        "      <menuitem action='Documentation'/>"
        "      <separator/>"
        "      <menuitem action='About'/>"
        "      <menuitem action='Stats'/>"
        "    </menu>"
        "  </menubar>"
        "</ui>";

#ifdef GLIBMM_EXCEPTIONS_ENABLED
    try {
        m_refUIManager->add_ui_from_string(ui_info);
    } catch(const Glib::Error& ex) {
        g_logger->error("building menus failed");
    }
#else
    std::auto_ptr<Glib::Error> ex;
    m_refUIManager->add_ui_from_string(ui_info, ex);
    if(ex.get()) {
        g_logger->error("building menus failed: " + ex->what());
    }
#endif
    pMenubar = m_refUIManager->get_widget("/MenuBar");

    m_Statusbar.set_homogeneous(0);
    m_Statusbar.pack_start(*g_logger->lbl_status_msg, Gtk::PACK_SHRINK);
    m_Statusbar.pack_end(*g_logger->lbl_status_upload, 0, 0);
    m_Statusbar.pack_end(*g_logger->lbl_status_download, 0, 0);

    m_ScrolledWindow.add(treeview);
    m_ScrolledWindow.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);

    m_Box.pack_start(*pMenubar, Gtk::PACK_SHRINK);
    m_Box.pack_start(m_ScrolledWindow);
    m_Box.pack_start(m_Statusbar, 0, 0);

    preferences_window = new Preferences;
    show_all();

    std::string buf;
    char c[512];

    snprintf(c, 512, "%s%cstats", g_settings->conf_dir.c_str(), DIR_SEP);
    bf::path stats_file(c);

    if(!bf::exists(stats_file)) {
        total_downloaded = 0;
        total_uploaded = 0;
    } else {
        int pos;
        std::fstream in(c, std::ios::in);
        while(in >> buf);
        in.close();

        pos = buf.find(':');

        total_downloaded = atoi(buf.substr(0, pos).c_str());
        total_uploaded = atoi(buf.substr(pos+1, buf.length()-pos).c_str());
    }

    /* Piping currently does not work on windows */
    this->pid_file = g_settings->pid_file.c_str();
    this->pipe_file = g_settings->pipe_file.c_str();
#ifndef WIN32
    pipe_create(this->pipe_file);

    FILE *fp = fopen(this->pid_file, "w+");
    if (fp) {
        fprintf(fp, "%d", getpid());
        fclose(fp);
    } else {
        g_logger->warning("could not create pid file");
    }
#endif
    elapsed = 0;
    refresh_rate = atoi(g_settings->get_option("refresh_rate"));
    if(refresh_rate < 50) refresh_rate = 50;
    Glib::signal_timeout().connect(sigc::mem_fun(*this, &MainWindow::on_timer), refresh_rate);

    g_logger->enable();

    if(treeview.loadTorrents()) {
        g_logger->message("Successfully loaded torrents");
    } else {
        g_logger->error("Could not load torrents!");
    }

    int i, sz = torrents.size();
    for (i = 0; i < sz; ++i) {
        treeview.add_torrent(torrents[i]);
    }
}

#ifndef WIN32
bool
MainWindow::pipe_create(const char *pipename)
{
    if (unlink(pipename) != 0 && errno != ENOENT) {
        g_logger->error("could not remove pipe");
        return false;
    }

    if (mkfifo(pipename, 0700) < 0) {
        g_logger->errorf("could not create named pipe %s", pipename);
        return false;
    }

    if ((this->fd = open(pipename, O_RDONLY | O_NONBLOCK)) < 0) {
        g_logger->errorf("could not create named pipe2 %s", pipename);
        return false;
    }

    chmod(pipename, 0707);

    g_logger->messagef("listening on pipe %s", pipename);

    return true;
}

bool
MainWindow::read_pipe()
{
    char *buf = NULL;
    int bufsize = 0;
    struct pollfd ps;
    ps.fd = this->fd;
    ps.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;

    if (poll(&ps, 1, 0) < 0) {
        g_logger->error("poll on pipe failed");
        return false;
    } else if ((ps.events & POLLNVAL)) {
        g_logger->error("got POLLNVAL on pipe, fd not open?");
        return false;
    } else if (ps.events & (POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI)) {
        char tmpbuf[8192U];
        int ret;

        ret = read(this->fd, tmpbuf, sizeof(tmpbuf));

        if (ret < 0) {
            if(errno == EAGAIN)
                return true;

            g_logger->error("read from pipe failed");
        } else if (ret == 0) {
            if (buf) {
#ifdef DEBUG
                g_logger->messagef("pipe buf: %s\n", buf);
#endif
                free(buf);
                buf = NULL;
                bufsize = 0;
            }

            return true;
        } else {
            while (ret > 0) {
                buf = (char*)realloc(buf, bufsize + ret + 1);
                memcpy(buf + bufsize, tmpbuf, ret);
                bufsize += ret;
                buf[bufsize-1] = 0x00;
                ret = read(fd, tmpbuf, sizeof(tmpbuf));
                std::string s(buf);
                unsigned int i = 0;
                size_t len = s.size();
                std::string tmp;
                for (i = 0; i <= len; i++) {
                    if (s[i] == '\n' || i == len) {
                        std::string tmp = s.substr(0, i);
                        unsigned int n = tmp.find(":");
                        std::string action = tmp.substr(0, n);
                        std::string arg = tmp.substr(n+1);

                        if (action == "add") {
                            /* check if there are any other arguments,
                             * like a specific directory */
                            n = arg.find(";");
                            if (n == std::string::npos) {
                                treeview.add_torrent(arg.c_str());
                            } else {
                                treeview.add_torrent(arg.substr(0, n).c_str(),
                                                     arg.substr(n+1));
                            }
                        } else if(action == "remove") {
                            treeview.remove_torrent(arg);
                        }

                        if (i == len) break;
                        
                        s = s.substr(i+1);
                        len = s.length();
                        i = 0;
                    }
                }
            }
        }
    }

    return true;
}
#endif

/** 
 * Save total downloaded/uploaded stats to CONF_DIR/stats
 **/
void
MainWindow::save_stats()
{
    std::string filename(g_settings->conf_dir+DIR_SEP+"stats");
    std::fstream out(filename.c_str(), std::ios::out);

    out << (g_torrent->get_total_downloaded()+total_downloaded) << ":"
        << (g_torrent->get_total_uploaded()+total_uploaded);

    out.close();
}

/**
 * Monitor a directory for newly added torrents
 **/
void
MainWindow::monitor_dir()
{
    bf::path p(g_settings->monitor_dir);

    if(!bf::exists(p)) {
        if(!bf::create_directory(p)) {
            g_logger->error("Could not create monitor directory", true);
            return;
        }
    }

    if(!bf::is_directory(p)) {
        g_logger->error("monitor directory is not a directory!");
        return;
    } else {
        bf::directory_iterator iter;
        for(bf::directory_iterator dir_iter(p); dir_iter != iter; ++dir_iter) {
#if BOOST_VERSION <= 103500
            if (bf::is_regular(dir_iter->status()))
#else
            if (bf::is_regular_file(dir_iter->status()))
#endif
            {
                bf::path file(dir_iter->string());
                if (treeview.add_torrent(dir_iter->string().c_str())
                 && bf::exists(file)) {
                    bf::remove(file);
                    g_logger->message("Successfully added "+dir_iter->string());
                }
            }
        }
    }
}

bool
MainWindow::on_timer()
{
    elapsed += refresh_rate;

    if (treeview.update_tooltip) {
        int torrents=0, completed_torrents=0;

        g_torrent->get_num_torrents(&torrents, &completed_torrents);

        sprintf(tooltip_cache, "torium - %d/%d", completed_torrents, torrents);

        treeview.update_tooltip = false;
        torrents_left = torrents-completed_torrents;

        if (torrents_left) {
            char tooltip[64];

            sprintf(tooltip, "%s\neta: %s", tooltip_cache,
                    fix_time(g_torrent->get_total_eta()).c_str());

            status_icon->set_tooltip(tooltip);
        } else {
            status_icon->set_tooltip(tooltip_cache);
        }
    }

    if (elapsed%5000 == 0) {
        /* Include total ETA remaining in the tooltip*/
        if (torrents_left) {
            char tooltip[64];

            sprintf(tooltip, "%s\neta: %s", tooltip_cache,
                    fix_time(g_torrent->get_total_eta()).c_str());

            status_icon->set_tooltip(tooltip);
        }
    }

    if (treeview.force_update) {
        save_torrents_status();
        
        treeview.force_update = false;
    }

    if (elapsed >= 60*1000) {
        elapsed = 0;

        save_torrents_status();
        save_stats();
        monitor_dir();
    }
#ifndef WIN32
    read_pipe();
#endif
    g_torrent->update_rates();

    /* Update the status bar with the most up-to-date information */
    g_logger->update_status(g_torrent->download_rate(), g_torrent->upload_rate());

    treeview.refreshTorrents();

    /* Handle all torrent alerts sent by libtorrent */
    treeview.check_torrent_alerts();
    return true;
}

void
MainWindow::on_status_icon_click()
{
    this->show();
    this->raise();
}

void
MainWindow::on_status_icon_popup(guint button, guint32 activate_time)
{
    status_icon_menu.popup(button, activate_time);
}

/** 
 * Save torrent statuses in ~/.torium/status
 **/
void
MainWindow::save_torrents_status()
{
    std::string filename(g_settings->conf_dir+DIR_SEP+"status");
    std::ofstream out(filename.c_str());
    std::vector<torrent_handle> torrent_list = g_torrent->get_torrents();
    int i, sz;

    /** 
     * File syntax
     *   unixtime_last_edited
     * followed by a list of all the torrents in this format
     *   \0torrent_name\0torrent_progress
     **/

    time_t sec;
    sec = time(NULL);
    out << sec;

    sz = torrent_list.size();
    for (i = 0; i < sz; ++i) {
        if (!torrent_list[i].is_valid()) {
            continue;
        }
        torrent_status s = torrent_list[i].status();
        int progress = 0;
        if (s.progress*100>0) {
            progress = (int)(s.progress*100);
        }

        out << '\n' << torrent_list[i].name() << '\0'
            << progress << '\0'
            << s.download_rate << '\0'
            << s.upload_rate;

    }

    out.close();
}

/** 
 * Function called when the Help menu item is clicked. 
 * Open up a web browser navigated to the project wiki
 * at bithack.se.
 **/
void
MainWindow::on_menu_help_contents()
{
    char *argv[] = {"xdg-open", "http://bithack.se/projects/torium/", 0};
    g_spawn_async(0, argv, 0, G_SPAWN_SEARCH_PATH, 0, 0, 0, 0);
}

/** 
 * Display a simple about dialog
 **/
void
MainWindow::on_menu_help_about() {
    std::vector<std::string> authors(2);
    authors[0] = "Rasmus Karlsson <pajlada@bithack.se>";
    authors[1] = "Emil Romanus <sdac@bithack.se>";
    Gtk::AboutDialog dialog;
    dialog.set_authors(authors);
    dialog.set_version(PACKAGE_VERSION);
    dialog.set_website("http://bithack.se/projects/torium");
    dialog.set_program_name("Torium");

    dialog.run();
}

/** 
 * Display a dialog displaying statistics
 **/
void
MainWindow::on_menu_help_stats() {
    Gtk::MessageDialog dialog(*this, "Stats");
    
    char c[512];
    long int tmp_dn   = g_torrent->get_total_downloaded();
    long int tmp_up   = g_torrent->get_total_uploaded();
    snprintf(c, 512, "<b>This session</b>\nDownloaded: %ldMb\nUploaded: %ldMb\n"
                     "<b>Total</b>\nDownloaded: %ldMb\nUploaded: %ldMb",
                     tmp_dn, tmp_up,
                     total_downloaded-tmp_dn, total_uploaded-tmp_up);

    dialog.set_secondary_text(c, true);
    dialog.run();
}

/**
 * Display the preferences dialog
 **/
void
MainWindow::on_menu_edit_preferences() {
    preferences_window->reset();
    preferences_window->show_all();
    preferences_window->raise();
}

/** 
 * Manually reload ipfilters
 **/
void
MainWindow::on_menu_edit_load_ip_filters() {
    treeview.load_ip_filters();
}

void
MainWindow::on_menu_file_add_torrent() {
    g_logger->message("Adding new torrent...");

    Gtk::FileChooserDialog dialog("Please choose a torrent file",
            Gtk::FILE_CHOOSER_ACTION_OPEN);
    dialog.set_transient_for(*this);
    dialog.set_select_multiple();

    dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
    dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

    Gtk::FileFilter filter_torrent;
    filter_torrent.set_name("Torrent files");
    filter_torrent.add_mime_type("application/x-bittorrent");
    dialog.add_filter(filter_torrent);

    Gtk::FileFilter filter_any;
    filter_any.set_name("Any files");
    filter_any.add_pattern("*");
    dialog.add_filter(filter_any);

    int result = dialog.run();

    switch(result) {
        case(Gtk::RESPONSE_OK):
        {
            std::vector<Glib::ustring> files = dialog.get_filenames();
            std::vector<Glib::ustring>::iterator i;
            for (i = files.begin(); i != files.end(); ++i)
                treeview.add_torrent((*i).c_str());

            break;
        }
    }
}

void
MainWindow::on_menu_file_quit() {
    g_logger->message("Quitting...");
    Gtk::Main::quit();
}

MainWindow::~MainWindow()
{
    int width, height,
        x, y;

    get_position(x, y);
    get_size(width, height);

    /* save window geometry for the next time the
     * user starts torium */
    g_settings->set_option("window_width", width);
    g_settings->set_option("window_height", height);
    g_settings->set_option("window_x", x);
    g_settings->set_option("window_y", y);

    unlink(this->pipe_file);
    unlink(this->pid_file);

    delete preferences_window;

    hide_all();
}

void
MainWindow::on_menu_file_create_torrent()
{
    gui_create_torrent window;
    Gtk::Main::run(window);
}
